@@ -5,18 +5,17 @@ from nerblackbox.modules.datasets.formatter.suc_formatter import SUCFormatter
 from nerblackbox.modules.datasets.formatter.conll2003_formatter import (
     CoNLL2003Formatter,
 )
+from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter
 
 
 class AutoFormatter:
     @staticmethod
-    def for_dataset(ner_dataset: str):
+    def for_dataset(ner_dataset: str) -> BaseFormatter:
         if ner_dataset == "swedish_ner_corpus":
-            _formatter = SwedishNerCorpusFormatter()
+            return SwedishNerCorpusFormatter()
         elif ner_dataset == "suc":
-            _formatter = SUCFormatter()
+            return SUCFormatter()
         elif ner_dataset == "conll2003":
-            _formatter = CoNLL2003Formatter()
+            return CoNLL2003Formatter()
         else:
             raise Exception(f"ner_dataset = {ner_dataset} unknown.")
-
-        return _formatter
